package org.insa.graphs.algorithm.shortestpath;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.List;
import java.util.Random;

import org.insa.graphs.algorithm.ArcInspector;
import org.insa.graphs.algorithm.ArcInspectorFactory;
import org.insa.graphs.model.Graph;
import org.insa.graphs.model.Node;
import org.insa.graphs.model.io.BinaryGraphReader;
import org.insa.graphs.model.io.GraphReader;

public class RunTimeTrial {
    final static int TEST_COUNT_PER_MAP = 30;

    /*
     * Runs TEST_COUNT_PER_MAP performance test iterations for every map and
     * outputs as CSV to stdout. Bellman is performed only for smaller maps.
     * When running with VSC, paths are relative to the BE_Graphes directory.
     */
    public static void main(String[] args) throws Exception {
        System.out.println("N,P,D,A,B");
        runTest("./Maps/midi-pyrenees.mapgr", false);
        runTest("./Maps/france.mapgr", false);
        runTest("./Maps/belgium.mapgr", false);
        runTest("./Maps/bretagne.mapgr", false);
        runTest("./Maps/bordeaux.mapgr", true);
        runTest("./Maps/carre.mapgr", true);
        runTest("./Maps/fractal.mapgr", true);
        runTest("./Maps/insa.mapgr", true);
        runTest("./Maps/toulouse.mapgr", true);
    }

    public static void runTest(String mapName, boolean runBellman) throws IOException {
        final Graph graph = readGraph(mapName);

        for (int i = 0; i < TEST_COUNT_PER_MAP; i++) {
            final ShortestPathData data = getRandomShortestPathData(graph);

            final ShortestPathSolution dijkstraSolution = new DijkstraAlgorithm(data).run();
            final ShortestPathSolution aStarSolution = new AStarAlgorithm(data).run();
            final ShortestPathSolution bellmanSolution = runBellman ? new BellmanFordAlgorithm(data).run() : null;

            final int pathSize = dijkstraSolution.getPath() == null ? -1 : dijkstraSolution.getPath().getArcs().size();

            System.out.println(
                  graph.size() + ","
                + pathSize + ","
                + dijkstraSolution.getSolvingTime().toMillis() + ","
                + aStarSolution.getSolvingTime().toMillis() + ","
                + (runBellman ? bellmanSolution.getSolvingTime().toMillis() : "")
                            );
        }
    }


    private static Node getRandomNode(Graph graph) {
        final Random rand = new Random();
        final int graph_size = graph.getNodes().size();
        return graph.get(Math.abs(rand.nextInt()) % graph_size);
    }

    private static ArcInspector getRandomArcFilter() {
        // final Random rand = new Random();
        final List<ArcInspector> filters = ArcInspectorFactory.getAllFilters();
        //return filters.get(Math.abs(rand.nextInt()) % filters.size());
        return filters.get(0); // all arcs
    }

    private static ShortestPathData getRandomShortestPathData(Graph graph) {
        final Node origin = getRandomNode(graph);
        final Node destination = getRandomNode(graph);
        final ArcInspector filter = getRandomArcFilter();
        return new ShortestPathData(graph, origin, destination, filter);
    }

    private static Graph readGraph(String mapName) throws IOException {
        // Visit these directory to see the list of available files on Commetud.
        // When running with VSC, paths are relative to the BE_Graphes directory.
        // Create a graph reader.
        final GraphReader reader = new BinaryGraphReader(
            new DataInputStream(new BufferedInputStream(new FileInputStream(mapName))));

        // Read the graph. X
        final Graph graph = reader.read();

        // Close resources
        reader.close();

        return graph;
    }
}
