package org.insa.graphs.algorithm.shortestpath;

import static org.junit.Assert.*;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;

import org.insa.graphs.algorithm.AbstractInputData.Mode;
import org.insa.graphs.algorithm.ArcInspector;
import org.insa.graphs.algorithm.ArcInspectorFactory;
import org.insa.graphs.model.Graph;
import org.insa.graphs.model.Node;
import org.insa.graphs.model.Path;
import org.insa.graphs.model.Point;
import org.insa.graphs.model.io.BinaryGraphReader;
import org.insa.graphs.model.io.GraphReader;
import org.insa.graphs.model.io.PathReader;
import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.runners.Parameterized.Parameter;

public abstract class ShortestPathAlgorithmTest {

    public static GraphReader reader;
    public static ArrayList<Graph> graph = new ArrayList<Graph>();
    public static PathReader pathReader;
    public static ArrayList<Path> path = new ArrayList<Path>();
    public static Random rand = new Random();

    protected abstract ShortestPathAlgorithm initAlgo(ShortestPathData data);

    @Parameter
    // Liste de cartes utilisées
    static ArrayList<String> Maps = new ArrayList<String>(Arrays.asList("../Maps/carre.mapgr",
                                                                            "../Maps/insa.mapgr",
                                                                            "../Maps/toulouse.mapgr",
                                                                            "../Maps/midi-pyrenees.mapgr",
                                                                            "../Maps/belgium.mapgr"
                                                                            ));

    @BeforeClass
    /* 
     * Ajoute toute les cartes dans l'attribut graph.
     * Ensuite, on testera des chemins dont qui sont disponibles dans custom_path.
     * Ces chemins ont été tracés avec l'algo BellmanFord et sont donc considérés corrects.
     * On ne crée pas les chemins dans un attribut, on va les "recréer" virtuellement à
     * partir de leur origine/destination et en appliquant BellmanFord.
     * Cela permet une meilleure adaptabilité du code.
    */
    public static void init() {
        try {
            // Create the map
            for (String mapName : Maps) {
                // Create a graph reader
                final GraphReader reader = new BinaryGraphReader(
                    new DataInputStream(new BufferedInputStream(new FileInputStream(mapName))));
                // Read the graph.
                graph.add(reader.read());
                // free resources
                reader.close();
                }
            }
        catch (FileNotFoundException e) {
            System.err.println("File not found: " + e.getMessage());
            fail("File not found: " + e.getMessage());
        }
        catch (IOException e ) {
            System.err.println("Error reading file: " + e.getMessage());
            fail("Error reading file: " + e.getMessage());
        }
    }

    
    /* Stratégie:
     * Chemins courts testés par comparaison avec Bellman.
     * Chemin longs testés par comparaison avec chemins déjà construits. Bellman trop long.
     */

    /*
     * Verifies that path is valid and matches the one found by the Bellman algo
     * Otherwise the algo in not optimal.
     */
    private void assertBellmanHasSameResult(Graph graph, Node origin, Node destination, ArcInspector arcFilter) {
        final ShortestPathData data = new ShortestPathData(graph, origin, destination, arcFilter);

        final ShortestPathAlgorithm algo = initAlgo(data);
        final ShortestPathSolution path = algo.doRun();

        final BellmanFordAlgorithm bellman = new BellmanFordAlgorithm(data);
        final ShortestPathSolution bell_path = bellman.doRun();

        
        if (bell_path.getPath() != null) {
            assert(path.getPath().isValid());
            if (bell_path.isFeasible()) {
                assert(path.isFeasible());
                if (data.getMode() == Mode.LENGTH) {
                    assert(Math.abs((float) algo.getCostPath() - path.getPath().getLength()) < 1);
                    assert(Math.abs(path.getPath().getLength() - bell_path.getPath().getLength()) < 1);
                }
                else {
                    if (!(arcFilter == ArcInspectorFactory.getAllFilters().get(4))) {
                        assert(Math.abs(algo.getCostPath() - path.getPath().getMinimumTravelTime()) < 1);
                    }
                    assert(Math.abs(path.getPath().getMinimumTravelTime() - bell_path.getPath().getMinimumTravelTime()) < 1);
                }
            }
            else {
                assert(!path.isFeasible());
            }
        }
        else {
            assert(path.getPath() == null);
            assert(!path.isFeasible());
        }
    }

    /*
     * Realizes n tests with the given graph informations.
     * The n tests all use random origin and destination nodes id.
     * The outputs of the tests are compared with the algorithm Bellman.
     * This is why this function shall be used only for graphs of little size
     * (in terms of nodes OR arcs as complexity of Bellman is : O(n*m)).
    */
    private void nTestsWithBellman(Graph graph, Node origin, Node destination, ArcInspector arcFilter, int numberOfNodes, int numberOfTests) {
        for (int i = 0 ; i < numberOfTests ; i++) {
            origin = graph.get(Math.abs(rand.nextInt()) % numberOfNodes);
            destination = graph.get(Math.abs(rand.nextInt()) % numberOfNodes);

            assertBellmanHasSameResult(graph, origin, destination, arcFilter);
        }
    }

    /*
     * Verifies that no path is found
     */
    private void assertNoPathFound(Graph graph, Node origin, Node destination, ArcInspector arcFilter) {
        ShortestPathData data = new ShortestPathData(graph, origin, destination, arcFilter);

        final ShortestPathAlgorithm algo = initAlgo(data);
        final ShortestPathSolution path = algo.doRun();
        
        assert(!path.isFeasible());
        assert(path.getPath() == null);
    }

    /* 
     * Cette fois-ci, Bellman est trop long à utiliser un grand nombre de fois.
     * On va donc utiliser quelques techniques pour se rassurer:
     * -Le cout en distance est au moins la distance à vol d'oiseau entre les noeuds.
     * -les sous-chemins d'un PCC sont des PCC. Sachant cela, on doit être sur
     * qu'en appliquant un second Dijkstra sur 2 noeuds aléatoires dans le path
     * renvoyé par le premier Dijkstra (hors origine et destination), on
     * obtient :
     * -un plus court chemin qui est le même que le sous chemin relevé (sous PCC).
     * -un plus court chemin de coût inférieur au chemin (premier dijkstra) dont il a été extrait.
     * On teste 'nbrePCC' PCC et 'nbreSousPCC' sous PCC.
    */
    public void TestsCheminsTropLongs(Graph graph, Node origin, Node destination, ArcInspector arcFilter, int nbrePCC, int nbreSousPCC) {
        int longueur_path = 0;
        Node noeudSelectionneOrigine = graph.get(0);
        Node noeudSelectionneDestination = graph.get(0);
        int size_graph = graph.size();

        for (int i = 0 ; i < nbrePCC ; i++) {
            origin = graph.get(Math.abs(rand.nextInt()) % size_graph);
            destination = graph.get(Math.abs(rand.nextInt()) % size_graph);
    
            ShortestPathData data = new ShortestPathData(graph, origin, destination, arcFilter);
        
            final ShortestPathAlgorithm algo = initAlgo(data);
            final ShortestPathSolution path = algo.doRun();

            if (path.getPath() != null) {
                assert(path.getPath().isValid());
                assert(path.isFeasible());
                double distanceFromOriginToDestination = 0;
                if (data.getMode() == Mode.LENGTH) {
                    assert((Math.abs(algo.getCostPath() - path.getPath().getLength())) < 1);
                    distanceFromOriginToDestination = Point.distance(origin.getPoint(), destination.getPoint());
                }
                else {
                    assert((Math.abs(algo.getCostPath() - path.getPath().getMinimumTravelTime())) < 1);
                    distanceFromOriginToDestination = Point.distance(origin.getPoint(), destination.getPoint()) / (1000 * graph.getGraphInformation().getMaximumSpeed());
                }
                

                // Distance trouvée supérieure ou égale (improbable voire impossible) au chemin à vol d'oiseau.
                assert(distanceFromOriginToDestination <= algo.getCostPath());

                for (int j = 0; j < nbreSousPCC ; j++) {
                    longueur_path = path.getPath().size();
                    int milieu_path = longueur_path / 2;
                    
                    int indiceOrigine = Math.abs(rand.nextInt()) % milieu_path;
                    int indiceDestination = milieu_path + Math.abs(rand.nextInt()) % milieu_path - 1;

                    noeudSelectionneOrigine = path.getPath().getArcs().get(indiceOrigine).getOrigin();
                    
                    noeudSelectionneDestination = path.getPath().getArcs().get(indiceDestination).getOrigin();
                    
                    ShortestPathData dataSousPCC = new ShortestPathData(graph, noeudSelectionneOrigine, noeudSelectionneDestination, arcFilter);

                    final ShortestPathAlgorithm algoSousPCC = initAlgo(dataSousPCC);
                    final ShortestPathSolution pathSousPCC = algoSousPCC.doRun();

                    if (path.getPath() != null) {
                        // Note: comme les chemins des arcs sont récupérés au même en droit entre pathSousPCC et path,
                        // on peut comparer directement les objets soit leurs adresses. Sinon, il aurait fallu comparer
                        // l'id des noeuds origines et destination du chemin.
                        for (int k = 0 ; k < pathSousPCC.getPath().size()-1; k++) {
                            assert(pathSousPCC.getPath().getArcs().get(k) == path.getPath().getArcs().get(indiceOrigine + k));
                        }
                        assert(pathSousPCC.getPath().isValid());
                        assert(pathSousPCC.isFeasible());
                        if (data.getMode() == Mode.LENGTH) {
                            assert((Math.abs(algoSousPCC.getCostPath() - pathSousPCC.getPath().getLength())) < 1.0);
                        }
                        else {
                            assert((Math.abs(algoSousPCC.getCostPath() - pathSousPCC.getPath().getMinimumTravelTime())) < 1.0);
                        }

                        // sous PCC de coût inférieur au PCC
                        assert(algoSousPCC.getCostPath() <= algo.getCostPath());
                    }
                }
            }
        }
    }


    @Test
    /*
     * Chemin nul sur carte carrée.
     * L'origine et la destination sont les mêmes (noeud 3).
     * Tous chemins permis
     */
    public void chemin_nul_CARRE() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(0);

        int indiceNoeud = Math.abs(rand.nextInt()) % myGraph.size();

        Node origin = myGraph.get(indiceNoeud);
        Node destination = myGraph.get(indiceNoeud);

        assertNoPathFound(myGraph, origin, destination, arcInspector);
    }

    @Test
    /*
     * Chemin inexistant sur la carte INSA.
     * Origine: 224
     * Destination: 814.
     * Les 2 noeuds font partie de deux composantes non connexes.
     * Tous chemins permis.
    */
    public void chemin_inexistant_INSA() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(1);
        Node origin = myGraph.get(224);
        Node destination = myGraph.get(814);

        assertNoPathFound(myGraph, origin, destination, arcInspector);
    }


    @Test
    /*
     * Test du mode à vélo facultatif.
     * On va vérifier que le chemin obtenu par vélo en dijkstra est bien le même,
     * que celui obtenu par vélo en bellman.
     * Il est possible qu'un chemin en vélo n'existe pas. Par exemple, si le seul
     * chemin pour accéder à la destination est de passer par l'autoroute, il n'est
     * accessible qu'en voiture.
     * Carte utilisée : ../Maps/toulouse.mapgr
     */
    public void chemin_velo_uniquement() {
        
        Graph myGraph = graph.get(2);
        // Filter: forBicyclesCustomT
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(4);
        Node origin = myGraph.get(0);
        Node destination = myGraph.get(0);
        int size_graph = myGraph.size();

        nTestsWithBellman(myGraph, origin, destination, arcInspector, size_graph, 50);

    }


    @Test
    /*
     * Permet de tester les chemins courts en distance.
     * Map: carre.mapgr
     * 2000 chemins testés
     * Tous chemins permis
     * Mode: LENGTH
     * Carte utilisée: ../Maps/carre.mapgr
    */
    public void cheminsCarreLENGTH() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0); // all arcs
        Graph myGraph = graph.get(0);

        Node origin = myGraph.get(0);
        Node destination = myGraph.get(0);

        int size_graph = myGraph.getNodes().size();

        nTestsWithBellman(myGraph, origin, destination, arcInspector, size_graph, 2000);
    }


    @Test
    /*
     * Permet de tester les chemins courts en temps.
     * Map: carre.mapgr
     * 500 chemins testés
     * Tous chemins permis
     * Mode: LENGTH
     * Carte utilisée: ../Maps/carre.mapgr
    */
    public void cheminsCarreTIME() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(2); // all arcs
        Graph myGraph = graph.get(0);

        Node origin = myGraph.get(0);
        Node destination = myGraph.get(0);

        int size_graph = myGraph.getNodes().size();

        nTestsWithBellman(myGraph, origin, destination, arcInspector, size_graph, 2000);
    }

    @Test
    /*
     * Permet de tester les chemins moyennement longs en distance (0 - 30 km).
     * On teste 100 chemins sur la carte de Toulouse (environ 1 min d'exécution)
     * Même si Bellman est de plus long à faire long à réaliser avec la taille du
     * graphe, ce test prend moins d'une minute donc on va quand même utiliser Bellman.
     * Par contre, dans le test sur la région midi_pyrenees / France qui arrive après,
     * on va être obligé de trouver une autre solution (principale difficulté).
     * Mode: LENGTH
     * Carte utilisée : ../Maps/toulouse.mapgr
     */
    public void cheminsToulouseLENGTH() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(2);
        Node origin = myGraph.get(0);
        Node destination = myGraph.get(0);

        int size_graph = myGraph.getNodes().size();

        nTestsWithBellman(myGraph, origin, destination, arcInspector, size_graph, 100);
    }

    @Test
    /*
     * Permet de tester les chemins moyennement longs en temps (0 - 2h).
     * On teste 100 chemins sur la carte de Toulouse (environ 1 min d'exécution)
     * Même si Bellman est de plus long à faire long à réaliser avec la taille du
     * graphe, ce test prend moins d'une minute donc on va quand même utiliser Bellman.
     * Par contre, dans le test sur la région midi_pyrenees / France qui arrive après,
     * on va être obligé de trouver une autre solution (principale difficulté).
     * Mode: TIME
     * Carte utilisée : ../Maps/toulouse.mapgr
     */
    public void cheminsToulouseTIME() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(2);
        Node origin = myGraph.get(0);
        Node destination = myGraph.get(0);

        int size_graph = myGraph.getNodes().size();

        nTestsWithBellman(myGraph, origin, destination, arcInspector, size_graph, 100);
    }

    @Test
    /*
     * On veut vérifier le bon fonctionnement des modes TIME et LENGTH.
     * On va faire 100 tests pour vérifier le bon fonctionnement sur la carte
     * de Toulouse
     * Pour cela, on va obtenir deux chemins avec l'algo de algo et vérifier
     * que:
     * -le chemin TIME est plus rapide en durée que le chemin LENGTH
     * -le chemin LENGTH est plus court en distance que le chemin TIME.
     * On prend un grand chemin pour être sur d'avoir une différence :
     * Mode: TIME/LENGTH
     * Carte utilisée : ../Maps/toulouse.mapgr
     */
    public void chemin_time_length_comparison() {
        Graph myGraph = graph.get(2);
        Node origin = myGraph.get(0);
        Node destination = myGraph.get(0);
        int size_graph = myGraph.size();

        // Filter: allArcsL
        ArcInspector arcInspectorL = ArcInspectorFactory.getAllFilters().get(0);
        // Filter: forCarsT
        ArcInspector arcInspectorT = ArcInspectorFactory.getAllFilters().get(2);

        for (int i = 0 ; i < 100 ; i++) {
            origin = myGraph.get(Math.abs(rand.nextInt()) % size_graph);
            destination = myGraph.get(Math.abs(rand.nextInt()) % size_graph);

            ShortestPathData data = new ShortestPathData(myGraph, origin, destination, arcInspectorL);
            
            // Algo in mode LENGTH
            ShortestPathAlgorithm algo_L = initAlgo(data);
            ShortestPathSolution path_L = algo_L.doRun();

            data = new ShortestPathData(myGraph, origin, destination, arcInspectorT);

            // Algo in mode TIME
            ShortestPathAlgorithm algo_T = initAlgo(data);
            ShortestPathSolution path_T = algo_T.doRun();

            if (path_L.getPath() != null) {
                assert(path_L.getPath().isValid());
                assert(path_L.isFeasible());
                assert(path_T.getPath().isValid());
                assert(path_T.isFeasible());

                assert((Math.abs(algo_L.getCostPath() - path_L.getPath().getLength()) < 1.0));
                assert((Math.abs(algo_T.getCostPath() - path_T.getPath().getMinimumTravelTime()) < 1.0));

                assert(path_L.getPath().getLength() <= path_T.getPath().getLength());
                assert(path_L.getPath().getMinimumTravelTime() >= path_T.getPath().getMinimumTravelTime());
            }
        }
    }

    @Test
    /*
     * Le principe du test est celui utilisé dans la méthode TestsCheminsTropLongues:
     * -On fait plusieurs PCC et pour chacun on extrait plusieurs chemins des PCC
     * et on s'assure que c'est un PCC en refaisant notre algo de l'origine à la
     * destination du chemin extrait. Les chemins extraits et sousPCC obtenus doivent
     * être les mêmes.
     * Entres autres, on vérifie aussi que les PCC sont de distance supérieures
     * aux distances à vol d'oiseau et que les sousPCC sont de longueur
     * inférieures aux PCC.
     * On teste 10 PCC et pour chaque PCC 25 sous PCC
     * Mode: LENGTH
     * Carte utilisée : ../Maps/midi_pyrenees.mapgr
     */
    public void cheminsMidiPyreneesLENGTH() {
        Graph myGraph = graph.get(3);
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Node origin = myGraph.get(0);
        Node destination = myGraph.get(0);
        
        TestsCheminsTropLongs(myGraph, origin, destination, arcInspector, 10, 25);
    }
    
    @Test
    /*
     * Le principe du test est celui utilisé dans la méthode TestsCheminsTropLongues:
     * -On fait plusieurs PCC et pour chacun on extrait plusieurs chemins des PCC
     * et on s'assure que c'est un PCC en refaisant notre algo de l'origine à la
     * destination du chemin extrait. Les chemins extraits et sousPCC obtenus doivent
     * être les mêmes.
     * Entres autres, on vérifie aussi que les PCC sont de distance supérieures
     * aux distances à vol d'oiseau et que les sousPCC sont de longueur
     * inférieures aux PCC.
     * On teste 10 PCC et pour chaque PCC 25 sous PCC
     * Mode: TIME
     * Carte utilisée : ../Maps/midi_pyrenees.mapgr
     */
    public void cheminMidiPyreneesTIME() {
        Graph myGraph = graph.get(3);
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(2);
        Node origin = myGraph.get(0);
        Node destination = myGraph.get(0);
        
        TestsCheminsTropLongs(myGraph, origin, destination, arcInspector, 10, 25);
    }

    @Test
    /*
     * Le principe du test est celui utilisé dans la méthode TestsCheminsTropLongues:
     * -On fait plusieurs PCC et pour chacun on extrait plusieurs chemins des PCC
     * et on s'assure que c'est un PCC en refaisant notre algo de l'origine à la
     * destination du chemin extrait. Les chemins extraits et sousPCC obtenus doivent
     * être les mêmes.
     * Entres autres, on vérifie aussi que les PCC sont de distance supérieures
     * aux distances à vol d'oiseau et que les sousPCC sont de longueur
     * inférieures aux PCC.
     * On teste 10 PCC et pour chaque PCC 25 sous PCC
     * Mode: TIME
     * Carte utilisée : ../Maps/belgium.mapgr
     */
    public void cheminBelgiqueLENGTH() {
        Graph myGraph = graph.get(4);
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Node origin = myGraph.get(0);
        Node destination = myGraph.get(0);
        
        TestsCheminsTropLongs(myGraph, origin, destination, arcInspector, 10, 25);
    }

    @Test
    /*
     * Le principe du test est celui utilisé dans la méthode TestsCheminsTropLongues:
     * -On fait plusieurs PCC et pour chacun on extrait plusieurs chemins des PCC
     * et on s'assure que c'est un PCC en refaisant notre algo de l'origine à la
     * destination du chemin extrait. Les chemins extraits et sousPCC obtenus doivent
     * être les mêmes.
     * Entres autres, on vérifie aussi que les PCC sont de distance supérieures
     * aux distances à vol d'oiseau et que les sousPCC sont de longueur
     * inférieures aux PCC.
     * On teste 10 PCC et pour chaque PCC 25 sous PCC
     * Mode: TIME
     * Carte utilisée : ../Maps/belgium.mapgr
     */
    public void cheminBelgiqueTIME() {
        Graph myGraph = graph.get(4);
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(2);
        Node origin = myGraph.get(0);
        Node destination = myGraph.get(0);
        
        TestsCheminsTropLongs(myGraph, origin, destination, arcInspector, 10, 25);
    }
}

